# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-07-01 14:38
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_fido', '0010_credential_id_not_null'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='authenticator',
            name='credential_data',
        ),
    ]
