module.exports = {
    globals: {
        "jest": true,
        "expect": true,
        "mockFn": true,
        "config": true,
        "afterEach": true,
        "beforeEach": true,
        "describe": true,
        "it": true,
        "test": true,
        "runs": true,
        "waitsFor": true,
        "pit": true,
        "require": true,
        "xdescribe": true,
        "xit": true
    }
}
