# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-05-11 08:57
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_fido', '0013_unique_user_label'),
    ]

    operations = [
        migrations.CreateModel(
            name='AuthenticatorMetadata',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('url', models.URLField()),
                ('identifier', models.TextField(unique=True)),
                ('metadata_entry', models.TextField()),
                ('detailed_metadata_entry', models.TextField()),
            ],
        ),
    ]
